from ..dataset import Dataset
import importlib
import os


class TabularDataset(Dataset):
    dataset_type = os.path.basename(__file__).split('.py')[0]

    def __init__(self, name, dataset_type=None):
        super().__init__(name, dataset_type)

    def get_path(self):
        storage_type = super().get_storage_type().lower()
        get_path = importlib.import_module("." + storage_type, package="dsioutilities.tabular").get_path
        return get_path(self.name)




import glob
import os
from pysparkutilities import ds_initializer
from pysparkutilities.spark_initializer import spark_initializer
from bdaserviceutils import get_args


def from_spark_to_pandas_df_using_disk(df, path='/tmp/tmp_csv_df'):

    df = df.to_pandas_on_spark()
    df.to_csv(path, header=True, num_files=1, mode="overwrite")

    extension = 'csv'
    os.chdir(path)
    result = glob.glob('*.{}'.format(extension))[0]

    return result

def init_spark():
    return spark_initializer("Ds-io", get_args(), additional_config=[('spark.sql.execution.arrow.enabled', True)])


def get_path(name):
    df = ds_initializer.load_dataset(sc=init_spark(), name=name, read_all=False)
    return from_spark_to_pandas_df_using_disk(df)




from alidaargparser import get_asset_property

def input_or_output(name):

    dir = get_asset_property(name, property="direction")
    
    # TODO remove if else when multidataset supported
    if dir is not None:
        return dir
    else:
        if name == "input_dataset" or name == "input-dataset":
            return "input"
        else: 
            return "output"



from abc import ABC, abstractmethod
from .utils import input_or_output, get_asset_property
import logging

class Dataset(ABC):

    def __new__(cls, *args, **kw):
        
        
        if 'dataset_type' in kw:
            dataset_type = kw['dataset_type'].lower()
        elif len(args) > 0:
            dataset_type = args[1]
        else:
            dataset_type = "tabular"

        # Create a map of all subclasses based on dataset_type property (present on each subclass)
        subclass_map = {subclass.dataset_type: subclass for subclass in cls.__subclasses__()}


        # Select the proper subclass based on
        subclass = subclass_map[dataset_type]
        instance = super(Dataset, subclass).__new__(subclass)
        return instance
    
    def __init__(self, name, dataset_type = None):
        self.name = name
        self.direction = input_or_output(self.name)
        super().__init__()

    @abstractmethod
    def get_path(self, local_path, remote_path):
        pass

    def get_storage_type(self):
        storage_type = get_asset_property(asset_name=self.name, property="storage_type")
        if storage_type is not None:
            return storage_type
        else:
            logging.warning("Running locally.")
            return "FileSystem" 



import s3fs
from minio import Minio
from ..utils import input_or_output, get_asset_property

def set_connection(access_key, secret_key, url):

    minio_access_key = access_key
    minio_secret_key = secret_key
    endpoint_url = url

    class S3FileSystemPatched(s3fs.S3FileSystem):
        def __init__(self, *k, **kw):
            
            # Fix for calling set connection as many times as needed. Another fix is to remove kw from init method.
            if "key" in kw:
                kw.pop("key")
                kw.pop("secret")
                kw.pop("client_kwargs")
            
            super(S3FileSystemPatched, self).__init__( *k,
                                                    key = minio_access_key,
                                                    secret = minio_secret_key,
                                                    client_kwargs={'endpoint_url': endpoint_url},
                                                    **kw)
            print('S3FileSystem is patched with url:  ' + endpoint_url)

    s3fs.S3FileSystem = S3FileSystemPatched


def minio_ls(address, access_key, secret_key, bucket_name, folder, extention, use_ssl=False):

    if folder[-1] != "/":
        folder = folder + "/"

    client = Minio(
        address,
        access_key=access_key,
        secret_key=secret_key,
        secure=use_ssl
    )
    objects = client.list_objects(bucket_name=bucket_name, prefix=folder)

    return [x._object_name for x in objects if extention in x._object_name[-len(extention):]]




def get_path(name):
    
    # Get credentials
    minio_url = get_asset_property(asset_name=name, property="minIO_URL")
    access_key = get_asset_property(asset_name=name, property="minIO_ACCESS_KEY")
    secret_key = get_asset_property(asset_name=name, property="minIO_SECRET_KEY")
    bucket_name = get_asset_property(asset_name=name, property="minio_bucket")
    
    # Set connection for pandas reading
    set_connection(access_key=access_key, secret_key=secret_key, url=minio_url)

    if input_or_output(name) == "input":
        use_ssl = get_asset_property(asset_name=name, property="use_ssl") if get_asset_property(asset_name=name, property="use_ssl") is not None else False
        # List all files in dataset folder
        files_list = minio_ls(address=minio_url.replace("http://", "").replace("https://", ""), 
                                access_key=access_key, 
                                secret_key=secret_key, 
                                bucket_name=bucket_name, 
                                folder=get_asset_property(name),
                                extention=".csv",
                                use_ssl=use_ssl)
        if len(files_list) > 1:
            return ["s3://" + bucket_name + "/" + x for x in files_list]
        elif len(files_list) == 1:
            return "s3://" + bucket_name + "/" + files_list[0]
        else:
            raise Exception("Dataset is empty!")

    elif input_or_output(name) == "output":
        return "s3://" + bucket_name + "/" + get_asset_property(name) + "/dataset.csv"
    else:
        raise Exception("Error: cannot establish if dataset is an input or output one! Are you sure the name is right?")



import setuptools

setuptools.setup(
    name="ds-io-utilities",
    version="0.0.16",
    author="Alida research team",
    author_email="engineering-alida-lab@eng.it",
    description="Utils for datasets IO operations in Alida.",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires = [
        "hdfs>=2.0.0",
        "bda-service-utils",
        "s3fs==2022.1.0",
        "minio==7.1.8",
        "alida-arg-parser>=0.0.21"
        ],
)



import s3fs
from minio import Minio
from ..utils import input_or_output, get_asset_property


def get_path(name):
    
    bucket_name = get_asset_property(asset_name=name, property="minio_bucket")

    if input_or_output(name) == "input":
        return "s3://" + bucket_name + "/" + get_asset_property(name)
    elif input_or_output(name) == "output":
        return "s3://" + bucket_name + "/" + get_asset_property(name)
    else:
        raise Exception("Error: cannot establish if dataset is an input or output one! Are you sure the name is right?")

def get_metadata(name):

    use_ssl = get_asset_property(asset_name=name, property="use_ssl") if get_asset_property(asset_name=name, property="use_ssl") is not None else False
    if use_ssl:
        raise Exception("Error: ssl for object storage non implemented")

    # Get credentials
    minio_url = get_asset_property(asset_name=name, property="minIO_URL")
    access_key = get_asset_property(asset_name=name, property="minIO_ACCESS_KEY")
    secret_key = get_asset_property(asset_name=name, property="minIO_SECRET_KEY")
    bucket_name = get_asset_property(asset_name=name, property="minio_bucket")

    metadata = {"type": "minio", 
                "url": minio_url,
                "access_key": access_key,
                "secret_key": secret_key,
                "bucket_name": bucket_name,
                "use_ssl": use_ssl}

    return metadata